from contextlib import contextmanager
import pandas
import tempfile
import yaml
import numpy
import random
from results_schema import Model
from sqlalchemy.orm import sessionmaker
import datetime
import os
from triage.storage import MettaCSVMatrixStore
from metta import metta_io as metta
from collections import OrderedDict


@contextmanager
def fake_metta(matrix_dict, metadata):
    """Stores matrix and metadata in a metta-data-like form

    Args:
    matrix_dict (dict) of form { columns: values }.
        Expects an entity_id to be present which it will use as the index
    metadata (dict). Any metadata that should be set

    Yields:
        tuple of filenames for matrix and metadata
    """
    matrix = pandas.DataFrame.from_dict(matrix_dict).set_index('entity_id')
    with tempfile.NamedTemporaryFile() as matrix_file:
        with tempfile.NamedTemporaryFile('w') as metadata_file:
            hdf = pandas.HDFStore(matrix_file.name)
            hdf.put('title', matrix, data_columns=True)
            matrix_file.seek(0)

            yaml.dump(metadata, metadata_file)
            metadata_file.seek(0)
            yield (matrix_file.name, metadata_file.name)


def fake_labels(length):
    return numpy.array([random.choice([True, False]) for i in range(0, length)])


class MockTrainedModel(object):
    def predict_proba(self, dataset):
        return numpy.random.rand(len(dataset), len(dataset))


def fake_trained_model(project_path, model_storage_engine, db_engine, train_matrix_uuid='efgh'):
    """Creates and stores a trivial trained model

    Args:
        project_path (string) a desired fs/s3 project path
        model_storage_engine (triage.storage.ModelStorageEngine)
        db_engine (sqlalchemy.engine)

    Returns:
        (int) model id for database retrieval
    """
    trained_model = MockTrainedModel()
    model_storage_engine.get_store('abcd').write(trained_model)
    session = sessionmaker(db_engine)()
    db_model = Model(model_hash='abcd', train_matrix_uuid=train_matrix_uuid)
    session.add(db_model)
    session.commit()
    return trained_model, db_model.model_id


def sample_metta_csv_diff_order(directory):
    """Stores matrix and metadata in a metta-data-like form

    The train and test matrices will have different column orders

    Args:
        directory (str)
    """
    train_dict = OrderedDict([
        ('entity_id', [1, 2]),
        ('k_feature', [0.5, 0.4]),
        ('m_feature', [0.4, 0.5]),
        ('label', [0, 1])
    ])
    train_matrix = pandas.DataFrame.from_dict(train_dict)
    train_metadata = {
        'beginning_of_time': datetime.date(2014, 1, 1),
        'end_time': datetime.date(2015, 1, 1),
        'matrix_id': 'train_matrix',
        'label_name': 'label',
        'label_window': '3month',
        'indices': ['entity_id'],
    }

    test_dict = OrderedDict([
        ('entity_id', [3, 4]),
        ('m_feature', [0.4, 0.5]),
        ('k_feature', [0.5, 0.4]),
        ('label', [0, 1])
    ])

    test_matrix = pandas.DataFrame.from_dict(test_dict)
    test_metadata = {
        'beginning_of_time': datetime.date(2015, 1, 1),
        'end_time': datetime.date(2016, 1, 1),
        'matrix_id': 'test_matrix',
        'label_name': 'label',
        'label_window': '3month',
        'indices': ['entity_id'],
    }

    train_uuid, test_uuid = metta.archive_train_test(
        train_config=train_metadata,
        df_train=train_matrix,
        test_config=test_metadata,
        df_test=test_matrix,
        directory=directory,
        format='csv'
    )

    train_store = MettaCSVMatrixStore(
        matrix_path=os.path.join(directory, '{}.csv'.format(train_uuid)),
        metadata_path=os.path.join(directory, '{}.yaml'.format(train_uuid))
    )
    test_store = MettaCSVMatrixStore(
        matrix_path=os.path.join(directory, '{}.csv'.format(test_uuid)),
        metadata_path=os.path.join(directory, '{}.yaml'.format(test_uuid))
    )
    return train_store, test_store



import logging

from triage.experiments import ExperimentBase


class SingleThreadedExperiment(ExperimentBase):
    def build_matrices(self):
        logging.info('Creating sparse states')
        self.generate_sparse_states()
        logging.info('Creating labels')
        self.generate_labels()
        logging.info('Creating feature aggregation tables')
        self.feature_generator.process_table_tasks(self.feature_aggregation_table_tasks)
        logging.info('Creating feature imputation tables')
        self.feature_generator.process_table_tasks(self.feature_imputation_table_tasks)
        logging.info('Building all matrices')
        self.planner.build_all_matrices(self.matrix_build_tasks)

    def catwalk(self):
        for split_num, split in enumerate(self.full_matrix_definitions):
            self.log_split(split_num, split)
            train_store = self.matrix_store(split['train_uuid'])
            if train_store.empty:
                logging.warning('''Train matrix for split %s was empty,
                no point in training this model. Skipping
                ''', split['train_uuid'])
                continue
            if len(train_store.labels().unique()) == 1:
                logging.warning('''Train Matrix for split %s had only one
                unique value, no point in training this model. Skipping
                ''', split['train_uuid'])
                continue

            logging.info('Training models')
            model_ids = self.trainer.train_models(
                grid_config=self.config['grid_config'],
                misc_db_parameters=dict(
                    test=False,
                    model_comment=self.config.get('model_comment', None),
                ),
                matrix_store=train_store
            )
            logging.info('Done training models')

            for split_def, test_uuid in zip(
                split['test_matrices'],
                split['test_uuids']
            ):
                as_of_times = split_def['as_of_times']
                logging.info(
                    'Testing and scoring as_of_times min: %s max: %s num: %s',
                    min(as_of_times),
                    max(as_of_times),
                    len(as_of_times)
                )
                test_store = self.matrix_store(test_uuid)
                if test_store.empty:
                    logging.warning('''Test matrix for uuid %s
                    was empty, no point in generating predictions. Skipping.
                    ''', test_uuid)
                    continue
                for model_id in model_ids:
                    logging.info('Testing model id %s', model_id)
                    predictions_proba = self.predictor.predict(
                        model_id,
                        test_store,
                        misc_db_parameters=dict(),
                        train_matrix_columns=train_store.columns(),
                    )

                    self.individual_importance_calculator\
                        .calculate_and_save_all_methods_and_dates(
                            model_id,
                            test_store
                        )

                    self.evaluator.evaluate(
                        predictions_proba=predictions_proba,
                        labels=test_store.labels(),
                        model_id=model_id,
                        evaluation_start_time=split_def['matrix_start_time'],
                        evaluation_end_time=split_def['matrix_end_time'],
                        example_frequency=split_def['example_frequency']
                    )



from datetime import datetime, timedelta
import os
from sqlalchemy import create_engine
from functools import partial
from tempfile import TemporaryDirectory
import testing.postgresql
from unittest.mock import Mock
from unittest import TestCase

from catwalk.db import ensure_db
from catwalk.storage import FSModelStorageEngine

from triage.experiments import\
    MultiCoreExperiment,\
    SingleThreadedExperiment,\
    CONFIG_VERSION


def populate_source_data(db_engine):
    complaints = [
        (1, '2010-10-01', 5),
        (1, '2011-10-01', 4),
        (1, '2011-11-01', 4),
        (1, '2011-12-01', 4),
        (1, '2012-02-01', 5),
        (1, '2012-10-01', 4),
        (1, '2013-10-01', 5),
        (2, '2010-10-01', 5),
        (2, '2011-10-01', 5),
        (2, '2011-11-01', 4),
        (2, '2011-12-01', 4),
        (2, '2012-02-01', 6),
        (2, '2012-10-01', 5),
        (2, '2013-10-01', 6),
        (3, '2010-10-01', 5),
        (3, '2011-10-01', 3),
        (3, '2011-11-01', 4),
        (3, '2011-12-01', 4),
        (3, '2012-02-01', 4),
        (3, '2012-10-01', 3),
        (3, '2013-10-01', 4),
    ]

    entity_zip_codes = [
        (1, '60120'),
        (2, '60123'),
        (3, '60123'),
    ]

    zip_code_events = [
        ('60120', '2012-10-01', 1),
        ('60123', '2012-10-01', 10),
    ]

    events = [
        (1, 1, '2011-01-01'),
        (1, 1, '2011-06-01'),
        (1, 1, '2011-09-01'),
        (1, 1, '2012-01-01'),
        (1, 1, '2012-01-10'),
        (1, 1, '2012-06-01'),
        (1, 1, '2013-01-01'),
        (1, 0, '2014-01-01'),
        (1, 1, '2015-01-01'),
        (2, 1, '2011-01-01'),
        (2, 1, '2011-06-01'),
        (2, 1, '2011-09-01'),
        (2, 1, '2012-01-01'),
        (2, 1, '2013-01-01'),
        (2, 1, '2014-01-01'),
        (2, 1, '2015-01-01'),
        (3, 0, '2011-01-01'),
        (3, 0, '2011-06-01'),
        (3, 0, '2011-09-01'),
        (3, 0, '2012-01-01'),
        (3, 0, '2013-01-01'),
        (3, 1, '2014-01-01'),
        (3, 0, '2015-01-01'),
    ]

    states = [
        (1, 'state_one', '2012-01-01', '2016-01-01'),
        (1, 'state_two', '2013-01-01', '2016-01-01'),
        (2, 'state_one', '2012-01-01', '2016-01-01'),
        (2, 'state_two', '2013-01-01', '2016-01-01'),
        (3, 'state_one', '2012-01-01', '2016-01-01'),
        (3, 'state_two', '2013-01-01', '2016-01-01'),
    ]

    db_engine.execute('''create table cat_complaints (
        entity_id int,
        as_of_date date,
        cat_sightings int
        )''')

    db_engine.execute('''create table entity_zip_codes (
        entity_id int,
        zip_code text
        )''')

    for entity_zip_code in entity_zip_codes:
        db_engine.execute(
            "insert into entity_zip_codes values (%s, %s)", entity_zip_code
        )

    db_engine.execute('''create table zip_code_events (
        zip_code text,
        as_of_date date,
        num_events int
    )''')
    for zip_code_event in zip_code_events:
        db_engine.execute(
            "insert into zip_code_events values (%s, %s, %s)", zip_code_event
        )

    for complaint in complaints:
        db_engine.execute(
            "insert into cat_complaints values (%s, %s, %s)",
            complaint
        )

    db_engine.execute('''create table events (
        entity_id int,
        outcome int,
        outcome_date date
    )''')

    for event in events:
        db_engine.execute(
            "insert into events values (%s, %s, %s)",
            event
        )

    db_engine.execute('''create table states (
        entity_id int,
        state text,
        start_time timestamp,
        end_time timestamp
    )''')

    for state in states:
        db_engine.execute(
            'insert into states values (%s, %s, %s, %s)',
            state
        )


def num_linked_evaluations(db_engine):
    num_evaluations = len([
        row for row in db_engine.execute('''
            select * from results.evaluations e
            join results.models using (model_id)
            join results.predictions p on (
                e.model_id = p.model_id and
                e.evaluation_start_time <= p.as_of_date and
                e.evaluation_end_time > p.as_of_date)
        ''')
    ])
    return num_evaluations


def sample_config():
    temporal_config = {
        'beginning_of_time': '2010-01-01',
        'modeling_start_time': '2011-01-01',
        'modeling_end_time': '2014-01-01',
        'update_window': '1y',
        'train_label_windows': ['6months'],
        'test_label_windows': ['6months'],
        'train_example_frequency': '1day',
        'test_example_frequency': '3months',
        'train_durations': ['6months'],
        'test_durations': ['1months'],
    }

    scoring_config = {
        'metric_groups': [
            {'metrics': ['precision@'], 'thresholds': {'top_n': [2]}}
        ]
    }

    grid_config = {
        'sklearn.linear_model.LogisticRegression': {
            'C': [0.00001, 0.0001],
            'penalty': ['l1', 'l2'],
            'random_state': [2193]
        }
    }

    feature_config = [{
        'prefix': 'entity_features',
        'from_obj': 'cat_complaints',
        'knowledge_date_column': 'as_of_date',
        'aggregates_imputation': {'all': {'type': 'constant', 'value': 0}},
        'aggregates': [{
            'quantity': 'cat_sightings',
            'metrics': ['count', 'avg'],
        }],
        'intervals': ['1y'],
        'groups': ['entity_id']
    }, {
        'prefix': 'zip_code_features',
        'from_obj': 'entity_zip_codes join zip_code_events using (zip_code)',
        'knowledge_date_column': 'as_of_date',
        'aggregates_imputation': {'all': {'type': 'constant', 'value': 0}},
        'aggregates': [{
            'quantity': 'num_events',
            'metrics': ['max', 'min'],
        }],
        'intervals': ['1y'],
        'groups': ['entity_id', 'zip_code']
    }]

    state_config = {
        'table_name': 'states',
        'state_filters': ['state_one or state_two'],
    }

    return {
        'config_version': CONFIG_VERSION,
        'events_table': 'events',
        'entity_column_name': 'entity_id',
        'model_comment': 'test2-final-final',
        'model_group_keys': ['label_name', 'label_type', 'custom_key'],
        'feature_aggregations': feature_config,
        'state_config': state_config,
        'temporal_config': temporal_config,
        'grid_config': grid_config,
        'scoring': scoring_config,
        'user_metadata': { 'custom_key': 'custom_value' },
        'individual_importance': {'n_ranks': 2}
    }


def simple_experiment_test(experiment_class):
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        populate_source_data(db_engine)
        with TemporaryDirectory() as temp_dir:
            experiment_class(
                config=sample_config(),
                db_engine=db_engine,
                model_storage_class=FSModelStorageEngine,
                project_path=os.path.join(temp_dir, 'inspections')
            ).run()

        # assert
        # 1. that model groups entries are present
        num_mgs = len([
            row for row in
            db_engine.execute('select * from results.model_groups')
        ])
        assert num_mgs > 0

        # 2. that model entries are present, and linked to model groups
        num_models = len([
            row for row in db_engine.execute('''
                select * from results.model_groups
                join results.models using (model_group_id)
                where model_comment = 'test2-final-final'
            ''')
        ])
        assert num_models > 0

        # 3. predictions, linked to models
        num_predictions = len([
            row for row in db_engine.execute('''
                select * from results.predictions
                join results.models using (model_id)''')
        ])
        assert num_predictions > 0

        # 4. evaluations linked to predictions linked to models
        num_evaluations = len([
            row for row in db_engine.execute('''
                select * from results.evaluations e
                join results.models using (model_id)
                join results.predictions p on (
                    e.model_id = p.model_id and
                    e.evaluation_start_time <= p.as_of_date and
                    e.evaluation_end_time > p.as_of_date)
            ''')
        ])
        assert num_evaluations > 0

        # 5. experiment
        num_experiments = len([
            row for row in db_engine.execute('select * from results.experiments')
        ])
        assert num_experiments == 1

        # 6. that models are linked to experiments
        num_models_with_experiment = len([
            row for row in db_engine.execute('''
                select * from results.experiments
                join results.models using (experiment_hash)
            ''')
        ])
        assert num_models == num_models_with_experiment

        # 7. that models have the train end date and label window
        results = [
            (model['train_end_time'], model['train_label_window'])
            for model in db_engine.execute('select * from results.models')
        ]
        assert sorted(set(results)) == [
            (datetime(2011, 7, 1), timedelta(180)),
            (datetime(2012, 7, 1), timedelta(180)),
            (datetime(2013, 7, 1), timedelta(180)),
        ]

        # 8. that the right number of individual importances are present
        individual_importances = [row for row in db_engine.execute('''
            select * from results.individual_importances
            join results.models using (model_id)
        ''')]
        assert len(individual_importances) == num_predictions * 2  # only 2 features


def test_singlethreaded_experiment():
    simple_experiment_test(SingleThreadedExperiment)


def test_multicore_experiment():
    simple_experiment_test(
        partial(MultiCoreExperiment, n_processes=2, n_db_processes=2)
    )


def restart_experiment_test(experiment_class):
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        populate_source_data(db_engine)
        temp_dir = TemporaryDirectory()
        try:
            experiment = experiment_class(
                config=sample_config(),
                db_engine=db_engine,
                model_storage_class=FSModelStorageEngine,
                project_path=os.path.join(temp_dir.name, 'inspections'),
            )

            experiment.run()

            evaluations = num_linked_evaluations(db_engine)
            assert evaluations > 0

            experiment = experiment_class(
                config=sample_config(),
                db_engine=db_engine,
                model_storage_class=FSModelStorageEngine,
                project_path=os.path.join(temp_dir.name, 'inspections'),
                replace=False
            )
            experiment.make_entity_date_table = Mock()
            experiment.run()
            assert not experiment.make_entity_date_table.called
        finally:
            temp_dir.cleanup()


def test_restart_singlethreaded_experiment():
    restart_experiment_test(SingleThreadedExperiment)


def test_restart_multicore_experiment():
    restart_experiment_test(
        partial(MultiCoreExperiment, n_processes=2, n_db_processes=2)
    )


def nostate_experiment_test(experiment_class):
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        populate_source_data(db_engine)
        experiment_config = sample_config()
        del experiment_config['state_config']
        with TemporaryDirectory() as temp_dir:
            experiment_class(
                config=experiment_config,
                db_engine=db_engine,
                model_storage_class=FSModelStorageEngine,
                project_path=os.path.join(temp_dir, 'inspections')
            ).run()

        # assert
        # 1. that model groups entries are present
        num_mgs = len([
            row for row in
            db_engine.execute('select * from results.model_groups')
        ])
        assert num_mgs > 0

        # 2. that model entries are present, and linked to model groups
        num_models = len([
            row for row in db_engine.execute('''
                select * from results.model_groups
                join results.models using (model_group_id)
                where model_comment = 'test2-final-final'
            ''')
        ])
        assert num_models > 0

        # 3. predictions, linked to models
        num_predictions = len([
            row for row in db_engine.execute('''
                select * from results.predictions
                join results.models using (model_id)''')
        ])
        assert num_predictions > 0

        # 4. evaluations linked to predictions linked to models
        num_evaluations = len([
            row for row in db_engine.execute('''
                select * from results.evaluations e
                join results.models using (model_id)
                join results.predictions p on (
                    e.model_id = p.model_id and
                    e.evaluation_start_time <= p.as_of_date and
                    e.evaluation_end_time > p.as_of_date)
            ''')
        ])
        assert num_evaluations > 0

        # 5. experiment
        num_experiments = len([
            row for row in db_engine.execute('select * from results.experiments')
        ])
        assert num_experiments == 1

        # 6. that models are linked to experiments
        num_models_with_experiment = len([
            row for row in db_engine.execute('''
                select * from results.experiments
                join results.models using (experiment_hash)
            ''')
        ])
        assert num_models == num_models_with_experiment

        # 7. that models have the train end date and label window
        results = [
            (model['train_end_time'], model['train_label_window'])
            for model in db_engine.execute('select * from results.models')
        ]
        assert sorted(set(results)) == [
            (datetime(2011, 7, 1), timedelta(180)),
            (datetime(2012, 7, 1), timedelta(180)),
            (datetime(2013, 7, 1), timedelta(180)),
        ]


def test_nostate_singlethreaded_experiment():
    nostate_experiment_test(SingleThreadedExperiment)


def test_nostate_multicore_experiment():
    nostate_experiment_test(
        partial(MultiCoreExperiment, n_processes=2, n_db_processes=2)
    )


class TestConfigVersion(TestCase):
    def test_load_if_right_version(self):
        experiment_config = sample_config()
        experiment_config['config_version'] = CONFIG_VERSION
        with testing.postgresql.Postgresql() as postgresql:
            db_engine = create_engine(postgresql.url())
            ensure_db(db_engine)
            experiment = SingleThreadedExperiment(
                config=experiment_config,
                db_engine=db_engine,
                model_storage_class=FSModelStorageEngine,
                project_path='inspections'
            )

        assert isinstance(experiment, SingleThreadedExperiment)

    def test_noload_if_wrong_version(self):
        experiment_config = sample_config()
        experiment_config['config_version'] = 'v0'
        with self.assertRaises(ValueError):
            SingleThreadedExperiment(
                config=experiment_config,
                db_engine=None,
                model_storage_class=FSModelStorageEngine,
                project_path='inspections'
            )



from catwalk.db import ensure_db
from timechop.timechop import Timechop
from architect.label_generators import BinaryLabelGenerator
from architect.features import \
    FeatureGenerator,\
    FeatureDictionaryCreator,\
    FeatureGroupCreator,\
    FeatureGroupMixer
from architect.state_table_generators import StateTableGenerator
from architect.planner import Planner
from catwalk.model_trainers import ModelTrainer
from catwalk.predictors import Predictor
from catwalk.individual_importance import IndividualImportanceCalculator
from catwalk.evaluation import ModelEvaluator
from catwalk.utils import save_experiment_and_get_hash
from catwalk.storage import CSVMatrixStore
import os
from datetime import datetime
from abc import ABCMeta, abstractmethod
from functools import partial
import logging
from triage.experiments import CONFIG_VERSION


def dt_from_str(dt_str):
    return datetime.strptime(dt_str, '%Y-%m-%d')


class ExperimentBase(object):
    """The Base class for all Experiments."""
    __metaclass__ = ABCMeta

    def __init__(
        self,
        config,
        db_engine,
        model_storage_class=None,
        project_path=None,
        replace=True,
    ):
        self._check_config_version(config)
        self.config = config
        self.db_engine = db_engine
        if model_storage_class:
            self.model_storage_engine =\
                model_storage_class(project_path=project_path)
        self.matrix_store_class = CSVMatrixStore # can't make this configurable yet until the Architect obeys
        self.project_path = project_path
        self.replace = replace
        ensure_db(self.db_engine)

        self.labels_table_name = 'labels'
        self.features_schema_name = 'features'
        if project_path:
            self.matrices_directory = os.path.join(self.project_path, 'matrices')
            if not os.path.exists(self.matrices_directory):
                os.makedirs(self.matrices_directory)
        self.experiment_hash = save_experiment_and_get_hash(
            self.config,
            self.db_engine
        )
        self._split_definitions = None
        self._matrix_build_tasks = None
        self._collate_aggregations = None
        self._feature_aggregation_table_tasks = None
        self._feature_imputation_table_tasks = None
        self._all_as_of_times = None
        self._master_feature_dictionary = None
        self.initialize_factories()
        self.initialize_components()

    def _check_config_version(self, config):
        if 'config_version' in config:
            config_version = config['config_version']
        else:
            logging.warning('config_version key not found in experiment config. Assuming v1, which may not be correct')
            config_version = 'v1'
        if config_version != CONFIG_VERSION:
            raise ValueError(
                '''Experiment config '{}'
                does not match current version '{}'.
                Will not run experiment.'''\
                .format(config_version, CONFIG_VERSION)
            )

    def initialize_factories(self):
        split_config = self.config['temporal_config']

        self.chopper_factory = partial(
            Timechop,
            beginning_of_time=dt_from_str(split_config['beginning_of_time']),
            modeling_start_time=dt_from_str(split_config['modeling_start_time']),
            modeling_end_time=dt_from_str(split_config['modeling_end_time']),
            update_window=split_config['update_window'],
            train_label_windows=split_config['train_label_windows'],
            test_label_windows=split_config['test_label_windows'],
            train_example_frequency=split_config['train_example_frequency'],
            test_example_frequency=split_config['test_example_frequency'],
            train_durations=split_config['train_durations'],
            test_durations=split_config['test_durations'],
        )

        self.state_table_generator_factory = partial(
            StateTableGenerator,
            experiment_hash=self.experiment_hash,
            dense_state_table=self.config.get('state_config', {})
            .get('table_name', None),
            events_table=self.config['events_table']
        )

        self.label_generator_factory = partial(
            BinaryLabelGenerator,
            events_table=self.config['events_table'],
        )

        self.feature_dictionary_creator_factory = partial(
            FeatureDictionaryCreator,
            features_schema_name=self.features_schema_name,
        )

        self.feature_generator_factory = partial(
            FeatureGenerator,
            features_schema_name=self.features_schema_name,
            replace=self.replace,
            beginning_of_time=split_config['beginning_of_time']
        )

        self.feature_group_creator_factory = partial(
            FeatureGroupCreator,
            self.config.get('feature_group_definition', {'all': [True]})
        )

        self.feature_group_mixer_factory = partial(
            FeatureGroupMixer,
            self.config.get('feature_group_strategies', ['all'])
        )

        self.planner_factory = partial(
            Planner,
            beginning_of_time=dt_from_str(split_config['beginning_of_time']),
            label_names=['outcome'],
            label_types=['binary'],
            db_config={
                'features_schema_name': self.features_schema_name,
                'labels_schema_name': 'public',
                'labels_table_name': self.labels_table_name,
                # TODO: have planner/builder take state table later on, so we
                # can grab it from the StateTableGenerator instead of
                # duplicating it here
                'sparse_state_table_name': 'tmp_sparse_states_{}'.format(self.experiment_hash),
            },
            matrix_directory=self.matrices_directory,
            states=self.config.get('state_config', {}).get('state_filters', []),
            user_metadata=self.config.get('user_metadata', {}),
            replace=self.replace
        )

        self.trainer_factory = partial(
            ModelTrainer,
            project_path=self.project_path,
            experiment_hash=self.experiment_hash,
            model_storage_engine=self.model_storage_engine,
            model_group_keys=self.config['model_group_keys'],
            replace=self.replace
        )

        self.predictor_factory = partial(
            Predictor,
            model_storage_engine=self.model_storage_engine,
            project_path=self.project_path,
            replace=self.replace
        )

        self.indiv_importance_factory = partial(
            IndividualImportanceCalculator,
            n_ranks=self.config.get('individual_importance', {}).get('n_ranks', 5),
            methods=self.config.get('individual_importance', {}).get('methods', ['uniform']),
            replace=self.replace
        )

        self.evaluator_factory = partial(
            ModelEvaluator,
            sort_seed=self.config['scoring'].get('sort_seed', None),
            metric_groups=self.config['scoring']['metric_groups'],
        )

    def initialize_components(self):
        self.chopper = self.chopper_factory()
        self.label_generator = self.label_generator_factory(db_engine=self.db_engine)
        self.state_table_generator = self.state_table_generator_factory(db_engine=self.db_engine)
        self.feature_generator = self.feature_generator_factory(db_engine=self.db_engine)
        self.feature_dictionary_creator = self.feature_dictionary_creator_factory(db_engine=self.db_engine)
        self.feature_group_creator = self.feature_group_creator_factory()
        self.feature_group_mixer = self.feature_group_mixer_factory()
        self.planner = self.planner_factory(engine=self.db_engine)
        self.trainer = self.trainer_factory(db_engine=self.db_engine)
        self.predictor = self.predictor_factory(db_engine=self.db_engine)
        self.individual_importance_calculator = self.indiv_importance_factory(db_engine=self.db_engine)
        self.evaluator = self.evaluator_factory(db_engine=self.db_engine)

    @property
    def split_definitions(self):
        """Temporal splits based on the experiment's configuration

        Returns: (dict) temporal splits

        Example:
        ```
        {
            'beginning_of_time': {datetime},
            'modeling_start_time': {datetime},
            'modeling_end_time': {datetime},
            'train_matrix': {
                'matrix_start_time': {datetime},
                'matrix_end_time': {datetime},
                'as_of_times': [list of {datetime}s]
            },
            'test_matrices': [list of matrix defs similar to train_matrix]
        }
        ```
        """
        if not self._split_definitions:
            self._split_definitions = self.chopper.chop_time()
            logging.info(
                'Computed and saved split definitions: %s',
                self._split_definitions
            )
        return self._split_definitions

    @property
    def all_as_of_times(self):
        """All 'as of times' in experiment config

        Used for label and feature generation.

        Returns: (list) of datetimes
        """
        if not self._all_as_of_times:
            all_as_of_times = []
            for split in self.split_definitions:
                all_as_of_times.extend(split['train_matrix']['as_of_times'])
                logging.info(
                    'Adding as_of_times from train matrix: %s',
                    split['train_matrix']['as_of_times']
                )
                for test_matrix in split['test_matrices']:
                    logging.info(
                        'Adding as_of_times from test matrix: %s',
                        test_matrix['as_of_times']
                    )
                    all_as_of_times.extend(test_matrix['as_of_times'])

            logging.info(
                'Computed %s total as_of_times for label and feature generation',
                len(all_as_of_times)
            )
            self._all_as_of_times = list(set(all_as_of_times))
            logging.info(
                'Computed %s distinct as_of_times for label and feature generation',
                len(self._all_as_of_times)
            )
        return self._all_as_of_times

    @property
    def collate_aggregations(self):
        """collate Aggregation objects used by this experiment.
        
        Returns: (list) of collate.Aggregation objects
        """
        if not self._collate_aggregations:
            logging.info('Creating collate aggregations')
            self._collate_aggregations = self.feature_generator.aggregations(
                feature_aggregation_config=self.config['feature_aggregations'],
                feature_dates=self.all_as_of_times,
                state_table=self.state_table_generator.sparse_table_name,
            )
        return self._collate_aggregations


    @property
    def feature_aggregation_table_tasks(self):
        """All feature table query tasks specified by this Experiment

        Returns: (dict) keys are group table names, values are themselves dicts, each with keys for different stages of table creation (prepare, inserts, finalize) and with values being lists of SQL commands
        """
        if not self._feature_aggregation_table_tasks:
            logging.info(
                'Calculating feature tasks for %s as_of_times',
                len(self.all_as_of_times)
            )
            self._feature_aggregation_table_tasks = self.feature_generator\
                .generate_all_table_tasks(self.collate_aggregations, task_type='aggregation')
        return self._feature_aggregation_table_tasks

    @property
    def feature_imputation_table_tasks(self):
        """All feature imputation query tasks specified by this Experiment

        Returns: (dict) keys are group table names, values are themselves dicts,
            each with keys for different stages of table creation (prepare, inserts, finalize)
            and with values being lists of SQL commands
        """
        if not self._feature_imputation_table_tasks:
            logging.info(
                'Calculating feature tasks for %s as_of_times',
                len(self.all_as_of_times)
            )
            self._feature_imputation_table_tasks = self.feature_generator\
                .generate_all_table_tasks(self.collate_aggregations, task_type='imputation')
        return self._feature_imputation_table_tasks

    @property
    def master_feature_dictionary(self):
        """All possible features found in the database. Not all features will necessarily end up in matrices

        Returns: (list) of dicts, keys being feature table names and values being lists of feature names
        """
        if not self._master_feature_dictionary:
            self._master_feature_dictionary = self.feature_dictionary_creator\
                .feature_dictionary(
                    feature_table_names=self.feature_imputation_table_tasks.keys(),
                    index_column_lookup=self.feature_generator.index_column_lookup(self.collate_aggregations)
                )
            logging.info(
                'Computed master feature dictionary: %s',
                self._master_feature_dictionary
            )

        return self._master_feature_dictionary

    @property
    def feature_dicts(self):
        """Feature dictionaries, representing the feature tables and columns configured in this experiment after computing feature groups.

        Returns: (list) of dicts, keys being feature table names and values being lists of feature names
        """

        return self.feature_group_mixer.generate(
            self.feature_group_creator.subsets(self.master_feature_dictionary)
        )

    @property
    def matrix_build_tasks(self):
        """Tasks for all matrices that need to be built as a part of this Experiment.

        Each task contains arguments understood by Architect.build_matrix

        Returns: (list) of dicts
        """
        if not self._matrix_build_tasks:
            updated_split_definitions, matrix_build_tasks =\
                self.planner.generate_plans(
                    self.split_definitions,
                    self.feature_dicts
                )
            self._full_matrix_definitions = updated_split_definitions
            self._matrix_build_tasks = matrix_build_tasks
        return self._matrix_build_tasks

    @property
    def full_matrix_definitions(self):
        """Full matrix definitions

        Returns: (list) temporal and feature information for each matrix
        """
        if not self._full_matrix_definitions:
            updated_split_definitions, matrix_build_tasks =\
                self.planner.generate_plans(
                    self.split_definitions,
                    self.feature_dicts
                )
            self._full_matrix_definitions = updated_split_definitions
            self._matrix_build_tasks = matrix_build_tasks
        return self._full_matrix_definitions

    @property
    def all_label_windows(self):
        """All train and test label windows
        
        Returns: (list) label windows, in string form as they appeared in the experiment config
        """
        return list(set(
            self.config['temporal_config']['train_label_windows'] + \
            self.config['temporal_config']['test_label_windows']
        ))

    def generate_labels(self):
        """Generate labels based on experiment configuration

        Results are stored in the database, not returned
        """
        self.label_generator.generate_all_labels(
            self.labels_table_name,
            self.all_as_of_times,
            self.all_label_windows
        )

    def generate_sparse_states(self):
        self.state_table_generator.generate_sparse_table(
            as_of_dates=self.all_as_of_times
        )

    def update_split_definitions(self, new_split_definitions):
        """Update split definitions

        Args: (dict) split definitions (should have matrix uuids)
        """
        self._split_definitions = new_split_definitions

    def log_split(self, split_num, split):
        logging.info(
            'Starting train/test for %s out of %s: train range: %s to %s',
            split_num+1,
            len(self.full_matrix_definitions),
            split['train_matrix']['matrix_start_time'],
            split['train_matrix']['matrix_end_time'],
        )

    def matrix_store(self, matrix_uuid):
        """Construct a matrix store for a given matrix uuid, using the Experiment's #matrix_store_class

        Args:
            matrix_uuid (string) A uuid for a matrix
        """
        matrix_store = self.matrix_store_class(
            matrix_path=os.path.join(
                self.matrices_directory,
                '{}.csv'.format(matrix_uuid)
            ),
            metadata_path=os.path.join(
                self.matrices_directory,
                '{}.yaml'.format(matrix_uuid)
            )
        )
        return matrix_store

    @abstractmethod
    def build_matrices(self):
        """Generate labels, features, and matrices"""
        pass

    @abstractmethod
    def catwalk(self):
        """Train, test, and evaluate models"""
        pass

    def run(self):
        try:
            self.build_matrices()
        finally:
            logging.info('Matrix build done or errored, cleaning up state table')
            self.state_table_generator.clean_up()
        self.catwalk()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('LICENSE') as license_file:
    license = license_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.readlines()

with open('requirements_dev.txt') as dev_requirements_file:
    test_requirements = requirements + [
        line for line in dev_requirements_file.readlines()
        if '-r requirements' not in line
    ]

setup(
    name='triage',
    version='1.1.1',
    description="Risk modeling and prediction",
    long_description=readme,
    author="Center for Data Science and Public Policy",
    author_email='datascifellows@gmail.com',
    url='https://github.com/dssg/triage',
    packages=[
        'triage',
        'triage.experiments',
    ],
    package_dir={'triage':
                 'triage'},
    include_package_data=True,
    install_requires=requirements,
    license=license,
    zip_safe=False,
    keywords='triage',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



from catwalk.storage import InMemoryModelStorageEngine
from sqlalchemy import create_engine
from triage.experiments import ExperimentBase
import logging
from multiprocessing import Pool
from functools import partial
from catwalk.utils import Batch
import traceback


class MultiCoreExperiment(ExperimentBase):
    def __init__(self, n_processes=1, n_db_processes=1, *args, **kwargs):
        super(MultiCoreExperiment, self).__init__(*args, **kwargs)
        self.n_processes = n_processes
        self.n_db_processes = n_db_processes
        if kwargs['model_storage_class'] == InMemoryModelStorageEngine:
            raise ValueError('''
                InMemoryModelStorageEngine not compatible with MultiCoreExperiment
            ''')

    def catwalk(self):
        for split_num, split in enumerate(self.full_matrix_definitions):
            self.log_split(split_num, split)
            train_store = self.matrix_store(split['train_uuid'])
            logging.info('Checking out train matrix')
            if train_store.empty:
                logging.warning('''Train matrix for split %s was empty,
                no point in training this model. Skipping
                ''', split['train_uuid'])
                continue
            logging.info('Checking out train labels')
            if len(train_store.labels().unique()) == 1:
                logging.warning('''Train Matrix for split %s had only one
                unique value, no point in training this model. Skipping
                ''', split['train_uuid'])
                continue

            logging.info('Training models')
            trainer_tasks = self.trainer.generate_train_tasks(
                grid_config=self.config['grid_config'],
                misc_db_parameters=dict(
                    test=False,
                    model_comment=self.config.get('model_comment', None),
                ),
                matrix_store=train_store
            )
            partial_train_models = partial(
                train_model,
                trainer_factory=self.trainer_factory,
                db_connection_string=self.db_engine.url
            )
            logging.info(
                'Starting parallel training: %s tasks, %s processes',
                len(trainer_tasks),
                self.n_processes
            )
            model_ids = []
            for batch_model_ids in self.parallelize(
                partial_train_models,
                trainer_tasks,
                self.n_processes
            ):
                model_ids += batch_model_ids
            logging.info('Done training models')

            for split_def, test_uuid in zip(
                split['test_matrices'],
                split['test_uuids']
            ):
                as_of_times = split_def['as_of_times']
                logging.info(
                    'Testing and scoring as_of_times min: %s max: %s num: %s',
                    min(as_of_times),
                    max(as_of_times),
                    len(as_of_times)
                )
                test_store = self.matrix_store(test_uuid)
                if test_store.empty:
                    logging.warning('''Test matrix for train uuid %s
                    was empty, no point in training this model. Skipping
                    ''', split['train_uuid'])
                    continue
                partial_test_and_evaluate = partial(
                    test_and_evaluate,
                    predictor_factory=self.predictor_factory,
                    evaluator_factory=self.evaluator_factory,
                    indiv_importance_factory=self.indiv_importance_factory,
                    test_store=test_store,
                    db_connection_string=self.db_engine.url,
                    split_def=split_def,
                    train_matrix_columns=train_store.columns(),
                    config=self.config
                )
                logging.info(
                    'Starting parallel testing with %s processes',
                    self.n_db_processes
                )
                self.parallelize_with_success_count(
                    partial_test_and_evaluate,
                    model_ids,
                    self.n_db_processes
                )
                logging.info('Cleaned up concurrent pool')
            logging.info('Done with test matrix')
        logging.info('Done with split')

    def parallelize_with_success_count(
        self,
        partially_bound_function,
        tasks,
        n_processes,
        chunksize=1
    ):
        num_successes = 0
        num_failures = 0
        for successful in self.parallelize(
            partially_bound_function,
            tasks,
            n_processes,
            chunksize
        ):
            if successful:
                num_successes += 1
            else:
                num_failures += 1
        logging.info(
            'Done. successes: %s, failures: %s',
            num_successes,
            num_failures
        )

    def build_matrices(self):
        logging.info('Creating sparse states')
        self.generate_sparse_states()
        logging.info('Creating labels')
        self.generate_labels()
        logging.info(
            'Creating feature tables with %s processes',
            self.n_db_processes
        )
        for table_name, tasks in self.feature_aggregation_table_tasks.items():
            logging.info('Processing features for %s', table_name)
            self.feature_generator.run_commands(tasks.get('prepare', []))
            partial_insert = partial(
                insert_into_table,
                feature_generator_factory=self.feature_generator_factory,
                db_connection_string=self.db_engine.url
            )
            self.parallelize_with_success_count(
                partial_insert,
                tasks.get('inserts', []),
                n_processes=self.n_db_processes,
                chunksize=25
            )
            self.feature_generator.run_commands(tasks.get('finalize', []))
            logging.info('%s completed', table_name)

        logging.info('Creating feature imputation tables')
        self.feature_generator.process_table_tasks(self.feature_imputation_table_tasks)

        partial_build_matrix = partial(
            build_matrix,
            planner_factory=self.planner_factory,
            db_connection_string=self.db_engine.url
        )
        logging.info(
            'Starting parallel matrix building: %s matrices, %s processes',
            len(self.matrix_build_tasks.keys()),
            self.n_processes
        )
        self.parallelize_with_success_count(
            partial_build_matrix,
            self.matrix_build_tasks.values(),
            self.n_processes
        )

    def parallelize(
        self,
        partially_bound_function,
        tasks,
        n_processes,
        chunksize=1,
    ):
        with Pool(n_processes) as pool:
            for result in pool.map(
                partially_bound_function,
                [list(task_batch) for task_batch in Batch(tasks, chunksize)]
            ):
                yield result


def insert_into_table(
    insert_statements,
    feature_generator_factory,
    db_connection_string
):
    try:
        logging.info('Beginning insert batch')
        db_engine = create_engine(db_connection_string)
        feature_generator = feature_generator_factory(db_engine)
        feature_generator.run_commands(insert_statements)
        return True
    except Exception:
        logging.error('Child error: %s', traceback.format_exc())
        return False


def build_matrix(
    build_tasks,
    planner_factory,
    db_connection_string,
):
    try:
        db_engine = create_engine(db_connection_string)
        planner = planner_factory(engine=db_engine)
        for build_task in build_tasks:
            planner.build_matrix(**build_task)
        return True
    except Exception:
        logging.error('Child error: %s', traceback.format_exc())
        return False


def train_model(
    train_tasks,
    trainer_factory,
    db_connection_string,
):
    try:
        db_engine = create_engine(db_connection_string)
        trainer = trainer_factory(db_engine=db_engine)
        return [
            trainer.process_train_task(**train_task)
            for train_task in train_tasks
        ]
    except Exception:
        logging.error('Child error: %s', traceback.format_exc())
        return []


def test_and_evaluate(
    model_ids,
    predictor_factory,
    evaluator_factory,
    indiv_importance_factory,
    test_store,
    db_connection_string,
    split_def,
    train_matrix_columns,
    config
):
    try:
        db_engine = create_engine(db_connection_string)
        for model_id in model_ids:
            logging.info('Generating predictions for model id %s', model_id)
            predictor = predictor_factory(db_engine=db_engine)
            evaluator = evaluator_factory(db_engine=db_engine)
            individual_importance = indiv_importance_factory(db_engine=db_engine)

            predictions_proba = predictor.predict(
                model_id,
                test_store,
                misc_db_parameters=dict(),
                train_matrix_columns=train_matrix_columns
            )
            logging.info('Generating individual importances for model id %s', model_id)
            individual_importance.calculate_and_save_all_methods_and_dates(
                model_id,
                test_store
            )
            logging.info('Generating evaluations for model id %s', model_id)
            evaluator.evaluate(
                predictions_proba=predictions_proba,
                labels=test_store.labels(),
                model_id=model_id,
                evaluation_start_time=split_def['matrix_start_time'],
                evaluation_end_time=split_def['matrix_end_time'],
                example_frequency=split_def['example_frequency']
            )
        return True
    except Exception:
        logging.error('Child error: %s', traceback.format_exc())
        return False
